@@ -22,8 +22,8 @@
 // 1. split this file into several files, avoid virtual inheritance, generally
 //    improve this.
 
-#ifndef LIBSEMIGROUPS_INCLUDE_ELEMENT_HPP_
-#define LIBSEMIGROUPS_INCLUDE_ELEMENT_HPP_
+#ifndef LIBSEMIGROUPS_ELEMENT_HPP_
+#define LIBSEMIGROUPS_ELEMENT_HPP_
 
 #include <cstddef>  // for size_t
 #include <math.h>   // for sqrt, pow
@@ -1947,7 +1947,8 @@ namespace libsemigroups {
     }
   };
 
-  // FIXME why is there no specialisation for non-pointers for IncreaseDegree?
+  // FIXME(later) why is there no specialisation for non-pointers for
+  // IncreaseDegree?
 
   //! Specialization of the adapter Less for pointers to subclasses
   //! of Element.
@@ -2185,4 +2186,4 @@ namespace libsemigroups {
   };
 
 }  // namespace libsemigroups
-#endif  // LIBSEMIGROUPS_INCLUDE_ELEMENT_HPP_
+#endif  // LIBSEMIGROUPS_ELEMENT_HPP_
